/******************/
/* Initialization */


/* Updates the HUE bridge IP address and registration.
 * The address is ether autodetected, or set by the optional argument.
 *
 * @param hueBridgeIPAddress optional argument specifying a manual IP address
 *        if not set, or if set to an empty string, autodetection is used
 */
function initHUEBridge(hueBridgeIPAddress)
{
	try
	{
		lInfo.logln("initHUEBridge: updating HUE bridge address information");
		updateHueBridgeAddress(hueBridgeIPAddress);
		lInfo.logln("initHUEBridge: registering this application to the HUE bridge (if not done already)");
		hueBridgeRegisterApp();

		/* If this the first time a bridge was contacted successfully,
		 * set default presets and global activities */
		var registerStatus = getHueBridgeRegisterStatus();
		if (registerStatus === HueBridgeRegisterStatus.OK)
		{
			var hueBridgeContactedBefore = Property.getProperty(hueBridgeContactedBeforePropName);
			if (!hueBridgeContactedBefore)
			{
				createStandardGlobalActivities();
				createStandardPresetsForAllZones();

				Property.setProperty(hueBridgeContactedBeforePropName, true);
				Property.setFlag(hueBridgeContactedBeforePropName, 'ARCHIVE', true);
				Property.store();
			}
		}

		checkLightsOfAllActivities();
	}
	catch (err)
	{
		lWarn.logln("initHUEBridge: Error during HUE bridge processing: " + String(err));
	}
}


/* Initializes application. */
function startup()
{
	/* First, load the property tree; this must be the very first call,
	 * since most functions in the application make use of the property tree. */
	Property.load();

	/// Check if a manual IP address has been configured already
	var manual = htmlUnescape(Property.getProperty(hueBridgeAddrManualPropName));
	lInfo.logln("startup: loading HUE bridge IP address; manual value = " + String(manual));
	if (manual)
	{
		// Manual IP address given - initialize HUE bridge with it
		var ipaddress = htmlUnescape(Property.getProperty(hueBridgeAddrPropName));
		lInfo.logln("startup: using manually configured IP address " + String(ipaddress) + " for HUE bridge");
		initHUEBridge(ipaddress);
	}
	else
	{
		// no IP address given - let the Philips Portal API find the bridge
		lInfo.logln("startup: autodetecting IP address of HUE bridge");
		initHUEBridge();
	}
}




/*********************/
/* Activity handling */


/* Undo stack explanation:
 *
 * The undo "stack" is hardcoded to contain 2 levels. Level 1 is the "current" status, level 2
 * is the "previous" one. The current status is what the user just set, the previous the one before.
 * For each status, data about the associated trigger is stored: for scene triggers, this is the scene ID,
 * for statechange triggers, this is the state name.
 *
 * When storeCurBehaviorInUndoStack() is called, the contents of level 1 is stored in level 2,
 * and level 1 is filled with the new data from the function parameters. The function does a check
 * to see if the same trigger data is already present in level 1; if so, it does not do anything.
 * This makes sure repeated triggers do not cause the stack to be updated every time.
 *
 * When handleUndo() is called, the trigger data is extracted from the function parameters, and
 * compared with the one stored in level 2. If there is a mismatch, or if level 2 is empty, then
 * this function does nothing. Otherwise, it sets the lights to whatever state is stored in level 2.
 * This check makes sure no mismatching do-undo sequences occur. Example: first, "Doorbell" is triggered,
 * then "Alarm 1", then "Undo alarm 2". In such a case, the scene IDs will not match, and the code
 * will not revert to the stored values for "Doorbell".
 */


var UndoStackNodeName = "undoStack";
var UndoStackCurrentStatus = UndoStackNodeName + "/currentStatus";
var UndoStackPreviousStatus = UndoStackNodeName + "/previousStatus";
var UndoStackTriggerDataPropSuffix = "triggerData";
var UndoStackLightSettingsSuffix = "settings";


function storeCurBehaviorInUndoStack(triggerData, behavior)
{
	// go through the trigger data for the lights in the current status,
	// compare each light's trigger data with triggerData, and store
	// the IDs of the lights with different trigger data
	// (the corresponding previous status of these lights will be updated)
	var idsOfLightsToUpdate = [];
	for (var i = 0; i < behavior.length; ++i)
	{
		var light = behavior[i];

		if (light.blinking)
		{
			lInfo.logln("storeCurBehaviorInUndoStack: light with ID " + String(light.id) + " has blinking flag enabled -> NOT storing in undo stack");
			continue;
		}

		var curPropNameBase = UndoStackCurrentStatus + "/" + String(light.id);
		var curTriggerDataJSON = htmlUnescape(Property.getProperty(curPropNameBase + "/triggerData"));
		var curTriggerData = null;

		try
		{
			if (curTriggerDataJSON !== null)
				curTriggerData = JSON.parse(htmlUnescape(curTriggerDataJSON));
		}
		catch (e)
		{
			lWarn.logln("storeCurBehaviorInUndoStack: could not parse current trigger data of light with ID " + String(light.id) + " in undo stack");
			return;
		}

		var sameTriggerStoredAlready = true;
		if (curTriggerData !== null)
		{
			for (var v in curTriggerData)
			{
				if (triggerData[v] !== curTriggerData[v])
				{
					sameTriggerStoredAlready = false;
					break;
				}
			}
		}

		lInfo.logln("Light " + String(light.id) + " sameTriggerStoredAlready " + String(sameTriggerStoredAlready));

		if (!sameTriggerStoredAlready)
		{
			idsOfLightsToUpdate.push(light.id + "");
		}
	}

	// store current lights with different trigger data as previous ones
	for (var i = 0; i < idsOfLightsToUpdate.length; ++i)
	{
		var id = String(idsOfLightsToUpdate[i]);

		var curPropNameBase = UndoStackCurrentStatus + "/" + id;
		var curTriggerDataJSON = htmlUnescape(Property.getProperty(curPropNameBase + "/triggerData"));
		var curSettingsJSON = htmlUnescape(Property.getProperty(curPropNameBase + "/settings"));

		var prevPropNameBase = UndoStackPreviousStatus + "/" + id;
		Property.setProperty(prevPropNameBase + "/triggerData", curTriggerDataJSON);
		Property.setProperty(prevPropNameBase + "/settings", curSettingsJSON);
	}

	// store the trigger and light data as the current status
	var newTriggerDataJSON = JSON.stringify(triggerData);
	for (var i = 0; i < behavior.length; ++i)
	{
		var light = behavior[i];
		var curPropNameBase = UndoStackCurrentStatus + "/" + String(light.id);

		Property.setProperty(curPropNameBase + "/triggerData", newTriggerDataJSON);
		Property.setProperty(curPropNameBase + "/settings", JSON.stringify(light));
	}
}


function cleanupUndo()
{
	var undoStackNode = Property.getNode(UndoStackNodeName);
	var subnode = undoStackNode.getChild("currentStatus");
	if (subnode)
		undoStackNode.removeChild(subnode);
}


function handleUndo(parameter)
{
	// Process trigger parameters and check if the trigger type is supported
	var triggerData = {};
	if (parameter.sceneID !== undefined)
	{
		triggerData["type"] = "scene";
		triggerData["sceneID"] = parseInt(parameter.sceneID);
		if (!isGlobalOffSceneID(triggerData.sceneID))
		{
			triggerData["zoneID"] = parseInt(parameter.zoneID);
			triggerData["groupID"] = parseInt(parameter.groupID);
		}
	}
	else if (parameter.statename !== undefined)
	{
		triggerData["type"] = "statechange";
		// remove the "undo-" prefix from the name
		if (parameter.statename.substring(0, 5) == 'undo-')
			triggerData["name"] = parameter.statename.substring(5);
		else
			triggerData["name"] = parameter.statename;
	}
	else
	{
		lWarn.logln("handleUndo: unsupported trigger type");
		return;
	}

	var curStatusNode = Property.getNode(UndoStackCurrentStatus);
	if (!curStatusNode)
	{
		lInfo.logln("handleUndo: no current status to revert - not doing anything");
		return;
	}

	var lightSettings = [];

	/*
	 For each light:
	 1) Compare its trigger data to the one in triggerData. If there is a mismatch, skip that light. Otherwise, continue to (2)
	 2) Get the previous status. If there is no previous status node for this light, skip the light (there is no previous status to revert to). Otherwise, continue to (3)
	 3) Add the previous status to the lightSettings array
	 4) Copy the previous status to the current one
	 5) Delete the previous status node
	 */
	var lightNodes = curStatusNode.getChildren();
	for (var i = 0; i < lightNodes.length; ++i)
	{
		var curLightNode = lightNodes[i];

		// Get the current trigger data for the light
		var curTriggerDataJSON = htmlUnescape(Property.getProperty(curLightNode.getPath() + "/triggerData"));
		var curTriggerData = null;
		try
		{
			curTriggerData = JSON.parse(htmlUnescape(curTriggerDataJSON));
		}
		catch (err)
		{
			lWarn.logln("handleUndo: could not parse current trigger data of light with ID " + curLightNode.getName() + " in undo stack");
		}

		// Compare the trigger data to triggerData
		if (curTriggerData !== null)
		{
			var triggersAreDifferent = false;
			for (var v in triggerData)
			{
				if ((triggerData[v] === undefined) || (curTriggerData[v] === undefined) || (triggerData[v] !== curTriggerData[v]))
				{
					triggersAreDifferent = true;
					break;
				}
			}

			if (triggersAreDifferent)
			{
				lInfo.logln("handleUndo: current status of light with ID " + curLightNode.getName() + " has mismatching trigger data - not doing anything");
				continue;
			}
			else
				lInfo.logln("handleUndo: current status of light with ID " + curLightNode.getName() + " has matching trigger data - proceeding with undo");
		}
		else
		{
			lInfo.logln("handleUndo: no current status of light with ID " + curLightNode.getName() + " to revert - not doing anything");
			continue;
		}

		// Get the node for the light's previous status
		var prevLightNode = Property.getNode(UndoStackPreviousStatus + "/" + curLightNode.getName());
		if (prevLightNode === null)
		{
			lInfo.logln("handleUndo: there is no previous status node for light with ID " + curLightNode.getName() + " - skipping light");
			continue;
		}

		var prevTriggerDataJSON = htmlUnescape(Property.getProperty(prevLightNode.getPath() + "/triggerData"));
		var prevSettingsJSON = htmlUnescape(Property.getProperty(prevLightNode.getPath() + "/settings"));

		// Add the previous status to the lightSettings array
		{
			var prevSettings = null;
			try
			{
				prevSettings = JSON.parse(prevSettingsJSON);
			}
			catch (err)
			{
				lWarn.logln("handleUndo: could not parse previous settings of light with ID " + curLightNode.getName() + " in undo stack");
				continue;
			}
			lightSettings.push(prevSettings);
		}

		// Copy previous status to the current one
		{
			Property.setProperty(curLightNode.getPath() + "/triggerData", prevTriggerDataJSON);
			Property.setProperty(curLightNode.getPath() + "/settings", prevSettingsJSON);
		}

		// Remove the previous status node for this light
		{
			var prevStatusNode = Property.getNode(UndoStackPreviousStatus);
			var subnode = prevStatusNode.getChild(curLightNode.getName());
			if (subnode)
				prevStatusNode.removeChild(subnode);
		}
	}

	// Send the previous light settings to the bridge
	if (lightSettings.length > 0)
		sendLightsToHueBridge(lightSettings);
}


function storeParamDataInUndoStack(parameter, settings)
{
	if (parameter.sceneID !== undefined)
	{
		storeCurBehaviorInUndoStack(
			{
				type: 'scene',
				sceneID: parseInt(parameter.sceneID),
				zoneID: parseInt(parameter.zoneID),
				groupID: parseInt(parameter.groupID)
			},
			settings
		);
	}
	else if (parameter.statename !== undefined)
	{
		storeCurBehaviorInUndoStack(
			{
				type: 'statechange', name: parameter.statename
			},
			settings
		);
	}
}


function handleActivity(parameter)
{
	var propPath = parameter.path;

	var activated = htmlUnescape(Property.getProperty(propPath + "/activated"));
	/* NOT using an (!activated) check, since this property may not be set (this is the case for presets) */
	if (activated === false)
	{
		lInfo.logln('handleActivity: activity is deactivated');
		return;
	}

	var lightsJSON = htmlUnescape(Property.getProperty(propPath + "/lights"));

	if (lightsJSON === null)
	{
		lWarn.logln('handleActivity: found no lights field for activity at path "' + propPath + '"');
		return;
	}

	var lights = null;
	try
	{
		lights = JSON.parse(lightsJSON);
		lInfo.logln("handleActivity: sending light properties to HUE bridge: " + lightsJSON);
		sendLightsToHueBridge(lights);
	}
	catch (err)
	{
		lWarn.logln('handleActivity: error while parsing JSON data in lights field for activity at path "' + propPath + '": ' + err);
		return;
	}

	storeParamDataInUndoStack(parameter, lights);
}


function handleDimming(parameter)
{
	var zoneID = parseInt(parameter.zoneID);
	var sceneID = parseInt(parameter.sceneID);

	lInfo.logln("handleDimming: zone ID " + String(zoneID) + " scene ID " + String(sceneID));

	try
	{
		switch (sceneID)
		{
			case 11: hueDimLightsInZone(zoneID, -dimmingStepSize); break;
			case 12: hueDimLightsInZone(zoneID, +dimmingStepSize); break;
			case 13: hueDimLightsInZone(zoneID, -255); break;
			case 14: hueDimLightsInZone(zoneID, +255); break;
			default: break;
		}
	}
	catch(e)
	{
		lWarn.logln('handleDimming: dimming failed: ' + String(e));
	}
}


function handleOff(parameter)
{
	var zoneID = parseInt(parameter.zoneID);
	var sceneID = parseInt(parameter.sceneID);

	lInfo.logln("handleOff: zone ID " + String(zoneID) + " scene ID " + String(sceneID));

	try
	{
		var newSettings = hueSwitchOffLightsInZone(zoneID, sceneID);
		storeParamDataInUndoStack(parameter, newSettings);
	}
	catch(e)
	{
		lWarn.logln('handleOff: off failed: ' + String(e));
	}
}




/***********************/
/* main event handling */


function parsePayload(payloadEscaped)
{
	var payload = unescape(payloadEscaped);
	try
	{
		payload = JSON.parse(payload);
	}
	catch (err)
	{
		lWarn.logln("parsePayload: JSON parse error: " + String(err));
		payload = null;
	}

	return payload;
}


// NOTE: this function will modify the contents of the payload (it will add a "msgid" member)
function respondToBackendChannelEvent(receivedEvent, payload)
{
	if (payload)
		payload.msgid = receivedEvent.parameter.msgid;
	else
		payload = { msgid: receivedEvent.parameter.msgid };

	lInfo.logln("respondToBackendChannelEvent: payload: " + JSON.stringify(payload));

	try
	{
		var responseEvent = new Event("huecontroller.backendChannel", payload);
		responseEvent.raise();
	}
	catch (err)
	{
		lWarn.logln("respondToBackendChannelEvent: error while raising response event: " + String(err));
	}
}


function returnHueBridgeInfo(raisedEvent)
{
	var info = getHueBridgeInfo();
	var infoJSON = JSON.stringify(info);
	lInfo.logln("returnHueBridgeInfo: HUE bridge info: " + infoJSON);
	var registerStatus = getHueBridgeRegisterStatus();
	var errorMsg = '';
	if (registerStatus !== HueBridgeRegisterStatus.OK)
	{
		errorMsg = Property.getProperty(hueRegisterErrorMsgPropName);
		if (errorMsg === null)
			errorMsg = '';
	}

	// send response event
	respondToBackendChannelEvent(raisedEvent, { info: infoJSON, registerStatus: registerStatus, errorMsg: errorMsg });
}


// TODO: comment this out in release builds
//lInfo.logln("eventhandling: huecontroller event raised: " + JSON.stringify(raisedEvent));

try
{
	if (raisedEvent.name == 'hueController.handleTrigger')
	{
		handleActivity(raisedEvent.parameter);
	}
	else if (raisedEvent.name == 'hueController.handleDimming')
	{
//		handleDimming(raisedEvent.parameter);
	}
	else if (raisedEvent.name == 'hueController.handleOff')
	{
		handleOff(raisedEvent.parameter);
	}
	else if (raisedEvent.name == 'stateChange')
	{
		if (raisedEvent.parameter.state === 'inactive')
			handleUndo(raisedEvent.parameter);
	}
	else if (raisedEvent.name == 'undoScene')
	{
		handleUndo(raisedEvent.parameter);
	}
	else if (raisedEvent.name == 'running')
	{
		startup();
	}
	else if (raisedEvent.name == 'blinkingTimeout')
	{
		var blinkID = Property.getProperty(hueBlinkingIDPropName);
		lInfo.logln("eventhandling: blinking timeout, stored blinkID: " + String(blinkID) + " event blinkID: " + String(raisedEvent.parameter.blinkID));
		if (blinkID === parseInt(raisedEvent.parameter.blinkID))
		{
			lInfo.logln("eventhandling: timeout valid - setting lights to old value");
			var oldLightAttributesJSON = htmlUnescape(Property.getProperty(hueBlinkingAttributesPropName));

			try
			{
				var oldLightAttributes = JSON.parse(oldLightAttributesJSON);
				handleBlinkingTimeout(oldLightAttributes);
			}
			catch (e)
			{
				lWarn.logln("eventhandling: error while resetting lights after blinking: " + String(e));
			}

			Property.setProperty(hueBlinkingIDPropName, 0);
		}
		else
			lInfo.logln("eventhandling: timeout no longer valid - ignoring");
	}
	else if (raisedEvent.name == 'huecontroller')
	{
		/* activity events */
		if (raisedEvent.parameter.action == 'deleteActivity')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln('eventhandling: deleting activity of type "' + payload.activityType + '" with ID ' + String(payload.activityID));
				deleteActivity(payload.activityID, payload.activityType);
			}

			respondToBackendChannelEvent(raisedEvent);
		}
		else if (raisedEvent.parameter.action == 'testActivity')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln('eventhandling: test-running activity of type "' + payload.activityType + '" with ID ' + String(payload.activityID));
				testActivity(payload.activityID, payload.activityType);
			}
		}
		else if (raisedEvent.parameter.action == 'setActivity')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				if (payload.activityID === null)
					lInfo.logln('eventhandling: setting activity of type "' + payload.activityType + '" (new ID will be created)');
				else
					lInfo.logln('eventhandling: setting activity of type "' + payload.activityType + '" with ID ' + String(payload.activityID));
				lInfo.logln('eventhandling: payload = ' + JSON.stringify(payload)); // TODO: comment out in release builds
				setActivity(payload.activityID, payload.activityType, payload.activity);
			}

			respondToBackendChannelEvent(raisedEvent);
		}
		else if (raisedEvent.parameter.action == 'getActivity')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln('eventhandling: getting activity of type "' + payload.activityType + '" with ID ' + String(payload.activityID));

				var activity = getActivity(payload.activityID, payload.activityType);
				if (activity)
				{
					// send response event
					respondToBackendChannelEvent(raisedEvent, { activityID: payload.activityID, activityType: payload.activityType, activity: JSON.stringify(activity) });
				}
			}
		}
		else if (raisedEvent.parameter.action == 'getAllActivities')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln('eventhandling: returning all activities to frontend');
				var activities = getAllActivities(payload.activityType);
				var activitiesJSON = JSON.stringify(activities);

				// send response event
				respondToBackendChannelEvent(raisedEvent, { activityType: payload.activityType, activities: activitiesJSON });
			}
		}
		/* HUE bridge events */
		else if (raisedEvent.parameter.action == 'setHueBridgeAddress')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				var ipaddress = payload.ipaddress;
				if (ipaddress)
					lInfo.logln("eventhandling: setHueBridgeAddress called with manual IP address " + JSON.stringify(ipaddress));
				else
					lInfo.logln("eventhandling: setHueBridgeAddress called with autodetection requested");
				initHUEBridge(ipaddress);

				returnHueBridgeInfo(raisedEvent);
				refreshTriggers();
			}
		}
		else if (raisedEvent.parameter.action == 'findNewLights')
		{
			lInfo.logln("eventhandling: instructing HUE bridge to search for new lights");
			hueFindNewLights();
		}
		else if (raisedEvent.parameter.action == 'testLight')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln('eventhandling: setting test values for light with ID ' + String(payload.lightID));
				testLight(payload);
			}
		}
		else if (raisedEvent.parameter.action == 'getHueBridgeInfo')
		{
			lInfo.logln("eventhandling: getting HUE bridge information");
			returnHueBridgeInfo(raisedEvent);
		}
		else if (raisedEvent.parameter.action == 'registerHueBridge')
		{
			lInfo.logln("eventhandling: registering this application to the HUE bridge (if not done already)");
			hueBridgeRegisterApp();

			returnHueBridgeInfo(raisedEvent);
		}
		else if (raisedEvent.parameter.action == 'setHueLightZoneAssociation')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln("eventhandling: updating light<->zone association: " + JSON.stringify(payload));
				setHueLightZoneAssociation(payload.lightID, payload.zoneID);

				returnHueBridgeInfo(raisedEvent);
				refreshTriggers();
			}
		}
		else if (raisedEvent.parameter.action == 'getHueBridgeLightAttributes')
		{
			var payload = parsePayload(raisedEvent.parameter.payload);
			if (payload)
			{
				lInfo.logln("eventhandling: getting HUE light attributes for light of ID " + String(payload.lightID));
				var attributes = getHueBridgeLightAttributes(payload.lightID);
				var attributesJSON = JSON.stringify(attributes);
				lInfo.logln("eventhandling: light attributes: " + attributesJSON);

				// send response event

				// need to transmit attributes as JSON, because the event itself
				// cannot support too complex values (for example, objects containing other objects)
				respondToBackendChannelEvent(raisedEvent, { attributes: attributesJSON });
			}
		}
	}
}
catch (err)
{
	lWarn.logln("[ERROR] eventhandling: " + getExceptionObjectDetails(err));
}

